#include "itkImage.h"
#include "itkImageFileReader.h"
#include "itkImageFileWriter.h"
#include "itkImageRegionIterator.h"
#include "itkVector.h"
#include "stdio.h"
#include "stdlib.h"
#include "time.h"
#include "Matrix.h"

int main(int argc, char * argv[])
{
  if( argc < 2 || argc > 4 )
    {
    std::cerr << "Wrong argument!" << std::endl;
    return 1;
    }

  /////////////////////////////////////////////////////////
  //parameters
  float mu=0.2;
  int iteration=50;
  if(argc>2)
  {
	mu=atof(argv[2]);
  }
  if(argc==4)
  {
	iteration=atoi(argv[3]);
  }

  /////////////////////////////////////////////////////////
  typedef unsigned char PixelType;
  const unsigned int Dimension = 3;
  typedef itk::Image< PixelType, Dimension > ImageType;
  typedef itk::ImageRegionIterator< ImageType > IteratorType;
  typedef itk::ImageFileReader< ImageType > ReaderType;

  ReaderType::Pointer reader = ReaderType::New();
  const char * filename = argv[1];
  reader->SetFileName( filename );
  reader->Update();

  ImageType::Pointer image = reader->GetOutput();

  ImageType::RegionType region=image->GetLargestPossibleRegion();
  ImageType::SizeType size=region.GetSize();

  long width=size[0],height=size[1],frames=size[2];

  /////////////////////////////////////////////////////////
  //loading image data
  IteratorType it( image, region );

  Matrix f(height,width,frames);
  f.create();
  long n=0;
  for (it.GoToBegin();!it.IsAtEnd();++it,n++)
  {
	f.m[n]=(float)it.Get();
  }

  /////////////////////////////////////////////////////////
  //GVF initialization
  double beginTime,GVFTime;
  beginTime=(double)clock()/1000;

  f.normalize();

  Matrix fx(height,width,frames),fy(height,width,frames),fz(height,width,frames),u(height,width,frames),v(height,width,frames),o(height,width,frames),gradient(height,width,frames),laplacian(height,width,frames),temp(height,width,frames);

  fx=f.dx();
  fy=f.dy();
  fz=f.dz();

  u=fx;
  v=fy;
  o=fz;

  gradient.create();
  gradient.gradient(fx,fy,fz);

  fx.multiply2self(gradient);
  fy.multiply2self(gradient);
  fz.multiply2self(gradient);
  gradient.oneMinus2self();

  //from now on fx is gradient.*fx,fy is gradient.*fy,fz is gradient.*fz, and gradient is 1-gradient

  /////////////////////////////////////////////////////////
  //GVF iterations
  for(int i=0;i<iteration;i++)
  {

	  u.ensure();
	  v.ensure();
	  o.ensure();

	  temp=u.laplacian();
	  temp.multiply2self(mu*4);
	  u.multiply2self(gradient);
	  u.plus2self(temp);
	  u.plus2self(fx);
	  free(temp.m);

	  temp=v.laplacian();
	  temp.multiply2self(mu*4);
	  v.multiply2self(gradient);
	  v.plus2self(temp);
	  v.plus2self(fy);
	  free(temp.m);
	  
	  temp=o.laplacian();
	  temp.multiply2self(mu*4);
	  o.multiply2self(gradient);
	  o.plus2self(temp);
	  o.plus2self(fz);
	  free(temp.m);

  }

  GVFTime=(double)clock()/1000-beginTime;
  printf("\nThe GVF algorithm took %f seconds for %d iterations with mu=%f\n",GVFTime,iteration,mu);

  /////////////////////////////////////////////////////////
  //write vector file for u, v and o
  typedef float VectorPixelType;
  typedef itk::Image< VectorPixelType, Dimension > VectorImageType;
  typedef itk::ImageRegionIterator< VectorImageType > VectorIteratorType;
  VectorImageType::Pointer vectorImage = VectorImageType::New();

  VectorImageType::IndexType vectorStart;
  VectorImageType::SizeType  vectorSize;

  vectorSize[0]  = size[0];
  vectorSize[1]  = size[1];
  vectorSize[2]  = size[2];

  vectorStart[0] =   0;  
  vectorStart[1] =   0;  
  vectorStart[2] =   0;

  VectorImageType::RegionType vectorRegion;
  vectorRegion.SetSize( vectorSize );
  vectorRegion.SetIndex( vectorStart );
  
  // Pixel data is allocated
  vectorImage->SetRegions( vectorRegion );
  vectorImage->Allocate();

  VectorIteratorType vectorIt( vectorImage, vectorRegion );

  n=0;
  for (vectorIt.GoToBegin();!vectorIt.IsAtEnd();++vectorIt,n++)
  {
	vectorIt.Set(u.m[n]);
  }

  typedef itk::ImageFileWriter< VectorImageType > VectorWriterType;
  VectorWriterType::Pointer vectorWriter = VectorWriterType::New();

  vectorWriter->SetFileName( "u.mhd" );
  vectorWriter->SetInput(  vectorImage  );

  try
    {
    vectorWriter->Update();
    }
  catch( itk::ExceptionObject & exp ) 
    {
    std::cerr << "Exception caught !" << std::endl;
    std::cerr << exp << std::endl;
    }

  /////////////////////////////////////////////////////////

  n=0;
  for (vectorIt.GoToBegin();!vectorIt.IsAtEnd();++vectorIt,n++)
  {
	vectorIt.Set(v.m[n]);
  }
  vectorWriter->SetFileName( "v.mhd" );
  vectorWriter->SetInput(  vectorImage  );

  try
    {
    vectorWriter->Update();
    }
  catch( itk::ExceptionObject & exp ) 
    {
    std::cerr << "Exception caught !" << std::endl;
    std::cerr << exp << std::endl;
    }

  /////////////////////////////////////////////////////////

  n=0;
  for (vectorIt.GoToBegin();!vectorIt.IsAtEnd();++vectorIt,n++)
  {
	vectorIt.Set(o.m[n]);
  }
  vectorWriter->SetFileName( "o.mhd" );
  vectorWriter->SetInput(  vectorImage  );

  try
    {
    vectorWriter->Update();
    }
  catch( itk::ExceptionObject & exp ) 
    {
    std::cerr << "Exception caught !" << std::endl;
    std::cerr << exp << std::endl;
    }

  return 0;
}
