#include "itkImage.h"
#include "itkImageFileReader.h"
#include "itkImageFileWriter.h"
#include "itkImageRegionIterator.h"
#include "itkVector.h"
#include "stdio.h"
#include "stdlib.h"
#include "time.h"

extern "C"    // note that the functions defined in .cu files will use C linkage, thus this extern "C"
void CUDAgvf3D(float * f, float *u,float *v,float *o,int height,int width,int frames,float mu,int iteration,float minValue,float range);

int main(int argc, char * argv[])
{
  if( argc < 2 || argc > 4 )
    {
    std::cerr << "Wrong argument!" << std::endl;
    return 1;
    }

  /////////////////////////////////////////////////////////
  //parameters
  float mu=0.2;
  int iteration=50;
  if(argc>2)
  {
	mu=atof(argv[2]);
  }
  if(argc==4)
  {
	iteration=atoi(argv[3]);
  }

  /////////////////////////////////////////////////////////
  typedef unsigned char PixelType;
  const unsigned int Dimension = 3;
  typedef itk::Image< PixelType, Dimension > ImageType;
  typedef itk::ImageRegionIterator< ImageType > IteratorType;
  typedef itk::ImageFileReader< ImageType > ReaderType;

  ReaderType::Pointer reader = ReaderType::New();
  const char * filename = argv[1];
  reader->SetFileName( filename );
  reader->Update();

  ImageType::Pointer image = reader->GetOutput();

  ImageType::RegionType region=image->GetLargestPossibleRegion();
  ImageType::SizeType size=region.GetSize();

  int width=size[0],height=size[1],frames=size[2];
  /////////////////////////////////////////////////////////
  //loading image data
  IteratorType it( image, region );

  unsigned long imageSize=height*width*frames*sizeof(float);     
  float *f=(float *)malloc(imageSize);
  float *u=(float *)malloc(imageSize);
  float *v=(float *)malloc(imageSize);
  float *o=(float *)malloc(imageSize);
  float maxValue,minValue,range;


  long n=0,i=0,j=0,k=0;

  for (it.GoToBegin(),maxValue=(float)it.Get(),minValue=(float)it.Get();!it.IsAtEnd();++it,n++)
  {
	f[n]=(float)it.Get();
	if(f[n]>maxValue) maxValue=f[n];
	if(f[n]<minValue) minValue=f[n];
  }
  range=maxValue-minValue;

  //begin to record time
  double beginTime,GVFTime;
  beginTime=(double)clock()/1000;
  
  /////////////////////////////////////////////////////////
  //GVF
  CUDAgvf3D(f,u,v,o,height,width,frames,mu,iteration,minValue,range);

  GVFTime=(double)clock()/1000-beginTime;
  printf("\nThe GVF algorithm took %f seconds for %d iterations with mu=%f\n",GVFTime,iteration,mu);

  /////////////////////////////////////////////////////////
  //write vector file for u, v and o
  typedef float VectorPixelType;
  typedef itk::Image< VectorPixelType, Dimension > VectorImageType;
  typedef itk::ImageRegionIterator< VectorImageType > VectorIteratorType;
  VectorImageType::Pointer vectorImage = VectorImageType::New();

  VectorImageType::IndexType vectorStart;
  VectorImageType::SizeType  vectorSize;

  vectorSize[0]  = size[0];
  vectorSize[1]  = size[1];
  vectorSize[2]  = size[2];

  vectorStart[0] =   0;  
  vectorStart[1] =   0;  
  vectorStart[2] =   0;

  VectorImageType::RegionType vectorRegion;
  vectorRegion.SetSize( vectorSize );
  vectorRegion.SetIndex( vectorStart );
  
  // Pixel data is allocated
  vectorImage->SetRegions( vectorRegion );
  vectorImage->Allocate();

  VectorIteratorType vectorIt( vectorImage, vectorRegion );

  n=0;
  for (vectorIt.GoToBegin();!vectorIt.IsAtEnd();++vectorIt,n++)
  {
	vectorIt.Set(u[n]);
  }

  typedef itk::ImageFileWriter< VectorImageType > VectorWriterType;
  VectorWriterType::Pointer vectorWriter = VectorWriterType::New();

  vectorWriter->SetFileName( "u.mhd" );
  vectorWriter->SetInput(  vectorImage  );

  try
    {
    vectorWriter->Update();
    }
  catch( itk::ExceptionObject & exp ) 
    {
    std::cerr << "Exception caught !" << std::endl;
    std::cerr << exp << std::endl;
    }

  /////////////////////////////////////////////////////////

  n=0;
  for (vectorIt.GoToBegin();!vectorIt.IsAtEnd();++vectorIt,n++)
  {
	vectorIt.Set(v[n]);
  }
  vectorWriter->SetFileName( "v.mhd" );
  vectorWriter->SetInput(  vectorImage  );

  try
    {
    vectorWriter->Update();
    }
  catch( itk::ExceptionObject & exp ) 
    {
    std::cerr << "Exception caught !" << std::endl;
    std::cerr << exp << std::endl;
    }

  /////////////////////////////////////////////////////////

  n=0;
  for (vectorIt.GoToBegin();!vectorIt.IsAtEnd();++vectorIt,n++)
  {
	vectorIt.Set(o[n]);
  }
  vectorWriter->SetFileName( "o.mhd" );
  vectorWriter->SetInput(  vectorImage  );

  try
    {
    vectorWriter->Update();
    }
  catch( itk::ExceptionObject & exp ) 
    {
    std::cerr << "Exception caught !" << std::endl;
    std::cerr << exp << std::endl;
    }

  return 0;
}
